import os

import cv2
import h5py
import numpy as np
import tifffile as tiff


def read_volume(file, type='tif3d', key=None, start=0, stop=-1):
    """
    Reads a volume file/directory and returns the data in it as a numpy array
    :param file: path to the data
    :param type: type of the volume file (tif2d, tif3d, tifseq, hdf5, png or pngseq)
    :param key: key to the data (only necessary for hdf5 files)
    :param start: first slice to read (only necessary for sequences)
    :param stop: last slice (exclusive) to read (-1 means it will read up to the final slice, only necessary for sequences)
    :return: numpy array containing the data
    """

    if type == 'tif2d' or type == 'tif3d' or type == 'tif2D' or type == 'tif3D':
        volume = read_tif(file)
    elif type == 'tifseq':
        volume = read_tifseq(file, start=start, stop=stop)
    elif type == 'hdf5':
        volume = read_hdf5(file, key=key)
    elif type == 'png':
        volume = read_png(file)
    elif type == 'pngseq':
        volume = read_pngseq(file, start=start, stop=stop)
    else:
        volume = None

    return volume


def read_tif(file, dtype='uint8'):
    """
    Reads tif formatted file and returns the data in it as a numpy array
    :param file: path to the tif file
    :param dtype: data type of the numpy array
    :return: numpy array containing the data
    """

    data = tiff.imread(file).astype(dtype)

    return data


def read_tifseq(dir, dtype='uint8', start=0, stop=-1):
    """
    Read a sequence of 2D TIF files
    :param dir: directory that contains the files
    :param dtype: data type of the output
    :param start: first slice to read
    :param stop: last slice (exclusive) to read (-1 means it will read up to the final slice)
    """

    files = os.listdir(dir)
    files.sort()
    sz = tiff.imread(os.path.join(dir, files[0])).shape
    if stop < 0:
        stop = len(files)
    data = np.zeros((stop - start, sz[0], sz[1]), dtype=dtype)
    for i in range(start, stop):
        data[i-start] = tiff.imread(os.path.join(dir, files[i]))

    return data


def read_hdf5(file, dtype='uint8', key=None):
    """
    Reads an HDF5 file as a numpy array
    :param file: path to the hdf5 file
    :param dtype: data type of the numpy array
    :param key: key in the hfd5 file that provides access to the data
    :return: numpy array containing the data
    """
    f = h5py.File(file, 'r')
    data = np.array(f.get(key), dtype=dtype)
    f.close()

    return data


def read_png(file, dtype='uint8'):
    """
    Read a 2D PNG file
    :param file: file to be read
    :param dtype: data type of the output
    """

    data = cv2.imread(file, cv2.IMREAD_GRAYSCALE).astype(dtype)

    return data


def read_pngseq(dir, dtype='uint8', start=0, stop=-1):
    """
    Read a sequence of 2D PNG files
    :param dir: directory that contains the files
    :param dtype: data type of the output
    :param start: first slice to read
    :param stop: last slice (exclusive) to read (-1 means it will read up to the final slice)
    """

    files = os.listdir(dir)
    files.sort()
    sz = cv2.imread(os.path.join(dir, files[0]), cv2.IMREAD_GRAYSCALE).shape
    if stop < 0:
        stop = len(files)
    data = np.zeros((stop - start, sz[0], sz[1]), dtype=dtype)
    for i in range(start, stop):
        data[i-start] = cv2.imread(os.path.join(dir, files[i]), cv2.IMREAD_GRAYSCALE).astype(dtype)

    return data


def write_volume(data, file, type='tif3d', start=0, stop=-1, K=4):
    """
    Writes a numpy array to a volume file/directory
    :param data: 2D/3D numpy array
    :param file: path to the data
    :param type: type of the volume file (tif2d, tif3d, tifseq, png or pngseq)
    :param start: first slice to write (only necessary for sequences)
    :param stop: last slice (exclusive) to write (-1 means it will write up to the final slice, only necessary for sequences)
    :param K: length of the string index (optional, only for sequences)
    """

    if type == 'tif2d' or type == 'tif3d':
        write_tif(data, file)
    elif type == 'tifseq':
        write_tifseq(data, file, start=start, stop=stop, K=K)
    elif type == 'png':
        write_png(data, file)
    elif type == 'pngseq':
        write_pngseq(data, file, start=start, stop=stop, K=K)


def write_tif(x, file, dtype='uint8'):
    """
    Write a 2D/3D data set as a TIF file
    :param x: 2D/3D data array
    :param file: directory to write the data to
    :param dtype: data type of the output
    """

    tiff.imsave(file, x.astype(dtype))


def write_png(x, file):
    """
    Write a 2D data set to a PNG file
    :param x: 3D data array
    :param file: directory to write the data to
    """

    cv2.imwrite(file, x.astype('uint8'), [cv2.IMWRITE_PNG_COMPRESSION, 9])


def write_tifseq(x, dir, prefix='', start=0, stop=-1, dtype='uint8', K=4):
    """
    Write a 3D data set to a directory, slice by slice, as TIF files
    :param x: 3D data array
    :param dir: directory to write the data to
    :param prefix: prefix of the separate files
    :param start: first slice to write
    :param stop: last slice (exclusive) to write (-1 means it will read up to the final slice)
    :param dtype: data type of the output
    :param K: number of digits for the index
    """

    if not os.path.exists(dir):
        os.mkdir(dir)
    if stop < 0:
        stop = x.shape[0]
    for i in range(start, stop):
        i_str = num2str(i, K=K)
        tiff.imsave(dir + '/' + prefix + i_str + '.tif', (x[i, :, :]).astype(dtype))


def write_pngseq(x, dir, prefix='', start=0, stop=-1, K=4):
    """
    Write a 3D data set to a directory, slice by slice, as PNG files
    :param x: 3D data array
    :param dir: directory to write the data to
    :param prefix: prefix of the separate files
    :param start: first slice to write
    :param stop: last slice (exclusive) to write (-1 means it will read up to the final slice)
    :param K: number of digits for the index
    """

    if not os.path.exists(dir):
        os.mkdir(dir)
    if stop < 0:
        stop = x.shape[0]
    for i in range(start, stop):
        i_str = num2str(i, K=K)
        cv2.imwrite(dir + '/' + prefix + i_str + '.png', (x[i, :, :]).astype('uint8'),
                    [cv2.IMWRITE_PNG_COMPRESSION, 9])


def num2str(n, K=4):
    n_str = str(n)
    for k in range(0, K - len(n_str)):
        n_str = '0' + n_str
    return n_str



import copy
import random

import numpy as np
import torch

from neuralnets.util.io import read_volume


def sample_labeled_input(data, labels, input_shape, preloaded=True, type='pngseq', data_shape=None):
    """
    Generate an input and target sample of certain shape from a labeled dataset
    :param data: data to sample from (a 3D numpy array if preloaded, a directory containing the data else)
    :param labels: labels to sample from (a 3D numpy array if preloaded, a directory containing the data else)
    :param input_shape: (z, x, y) shape of the sample
    :param preloaded: boolean that specifies whether the data is already in RAM
    :param type: type of the dataset that should be loaded in RAM (only necessary if preloaded==False)
    :param data_shape: (z, x, y) shape of the dataset to sample from (only necessary if preloaded==False)
    :return: a random sample
    """
    # randomize seed
    np.random.seed()

    # extract input and target patch
    if preloaded:  # if preloaded, we can simply load it from RAM
        # generate random position
        z = np.random.randint(0, data.shape[0] - input_shape[0] + 1)
        x = np.random.randint(0, data.shape[1] - input_shape[1] + 1)
        y = np.random.randint(0, data.shape[2] - input_shape[2] + 1)

        input = data[z:z + input_shape[0], x:x + input_shape[1], y:y + input_shape[2]]
        target = labels[z:z + input_shape[0], x:x + input_shape[1], y:y + input_shape[2]]
    else:  # if not preloaded, we have to additionally load it in RAM
        # generate random position
        z = np.random.randint(0, data_shape[0] - input_shape[0] + 1)
        x = np.random.randint(0, data_shape[1] - input_shape[1] + 1)
        y = np.random.randint(0, data_shape[2] - input_shape[2] + 1)

        input = read_volume(data, type=type, start=z, stop=z + input_shape[0])
        target = read_volume(labels, type=type, start=z, stop=z + input_shape[0])
        input = input[:, x:x + input_shape[1], y:y + input_shape[2]]
        target = target[:, x:x + input_shape[1], y:y + input_shape[2]]

    return copy.copy(input), copy.copy(target)


def sample_unlabeled_input(data, input_shape, preloaded=True, type='pngseq', data_shape=None):
    """
    Generate an input sample of certain shape from an unlabeled dataset
    :param data: data to sample from (a 3D numpy array if preloaded, a directory containing the data else)
    :param input_shape: (z, x, y) shape of the sample
    :param preloaded: boolean that specifies whether the data is already in RAM
    :param type: type of the dataset that should be loaded in RAM (only necessary if preloaded==False)
    :param data_shape: (z, x, y) shape of the dataset to sample from (only necessary if preloaded==False)
    :return: a random sample
    """
    # randomize seed
    np.random.seed()

    # extract input and target patch
    if preloaded:  # if preloaded, we can simply load it from RAM
        # generate random position
        z = np.random.randint(0, data.shape[0] - input_shape[0] + 1)
        x = np.random.randint(0, data.shape[1] - input_shape[1] + 1)
        y = np.random.randint(0, data.shape[2] - input_shape[2] + 1)

        input = data[z:z + input_shape[0], x:x + input_shape[1], y:y + input_shape[2]]
    else:  # if not preloaded, we have to additionally load it in RAM
        # generate random position
        z = np.random.randint(0, data_shape[0] - input_shape[0] + 1)
        x = np.random.randint(0, data_shape[1] - input_shape[1] + 1)
        y = np.random.randint(0, data_shape[2] - input_shape[2] + 1)

        input = read_volume(data, type=type, start=z, stop=z + input_shape[0])
        input = input[:, x:x + input_shape[1], y:y + input_shape[2]]

    return copy.copy(input)


def gaussian_window(size, sigma=1):
    """
    Returns a 3D Gaussian window that can be used for window weighting and merging
    :param size: size of the window
    :param sigma: standard deviation of the gaussian
    :return: the Gaussian window
    """
    # half window sizes
    hwz = size[0] // 2
    hwy = size[1] // 2
    hwx = size[2] // 2

    # construct mesh grid
    if size[0] % 2 == 0:
        axz = np.arange(-hwz, hwz)
    else:
        axz = np.arange(-hwz, hwz + 1)
    if size[1] % 2 == 0:
        axy = np.arange(-hwy, hwy)
    else:
        axy = np.arange(-hwy, hwy + 1)
    if size[2] % 2 == 0:
        axx = np.arange(-hwx, hwx)
    else:
        axx = np.arange(-hwx, hwx + 1)
    xx, zz, yy = np.meshgrid(axx, axz, axy)

    # normal distribution
    gw = np.exp(-(xx ** 2 + yy ** 2 + zz ** 2) / (2. * sigma ** 2))

    # normalize so that the mask integrates to 1
    gw = gw / np.sum(gw)

    return gw


def load_net(model_file):
    """
    Load a pretrained pytorch network
    :param model_file: path to the checkpoint
    :return: a module that corresponds to the trained network
    """
    return torch.load(model_file)


def set_seed(seed):
    """
    Sets the seed of all randomized modules (useful for reproducibility)
    :param seed: seed number
    """
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)



import datetime
import os

import numpy as np
import torch
import torch.nn.functional as F
from progress.bar import Bar

from neuralnets.util.io import write_volume
from neuralnets.util.metrics import jaccard, dice, accuracy_metrics, hausdorff_distance
from neuralnets.util.tools import gaussian_window


def sliding_window_multichannel(image, step_size, window_size, track_progress=False):
    """
    Iterator that acts as a sliding window over a multichannel 3D image
    :param image: multichannel image (4D array)
    :param step_size: step size of the sliding window (3-tuple)
    :param window_size: size of the window (3-tuple)
    :param track_progress: optionally, for tracking progress with progress bar
    """

    # define range
    zrange = [0]
    while zrange[-1] < image.shape[1] - window_size[0]:
        zrange.append(zrange[-1] + step_size[0])
    zrange[-1] = image.shape[1] - window_size[0]
    yrange = [0]
    while yrange[-1] < image.shape[2] - window_size[1]:
        yrange.append(yrange[-1] + step_size[1])
    yrange[-1] = image.shape[2] - window_size[1]
    xrange = [0]
    while xrange[-1] < image.shape[3] - window_size[2]:
        xrange.append(xrange[-1] + step_size[2])
    xrange[-1] = image.shape[3] - window_size[2]

    # loop over the range
    if track_progress:
        bar = Bar('Progress', max=len(zrange) * len(yrange) * len(xrange))
    for z in zrange:
        for y in yrange:
            for x in xrange:

                # yield the current window
                if window_size[0] == 1:  # 2D
                    yield (z, y, x, image[:, z, y:y + window_size[1], x:x + window_size[2]])
                else:  # 3D
                    yield (z, y, x, image[:, z:z + window_size[0], y:y + window_size[1], x:x + window_size[2]])
                if track_progress:
                    bar.next()
    if track_progress:
        bar.finish()


def segment_multichannel(data, net, input_shape, batch_size=1, step_size=None, train=False, track_progress=False):
    """
    Segment a multichannel 3D image using a specific network
    :param data: 4D array (C, Z, Y, X) representing the multichannel 3D image
    :param net: image-to-image segmentation network
    :param input_shape: size of the inputs (either 2 or 3-tuple)
    :param batch_size: batch size for processing
    :param step_size: step size of the sliding window
    :param train: evaluate the network in training mode
    :param track_progress: optionally, for tracking progress with progress bar
    :return: the segmented image
    """

    # make sure we compute everything on the gpu and in the correct mode
    if torch.cuda.is_available():
        net.cuda()
    else:
        net.cpu()
    if train:
        net.train()
    else:
        net.eval()

    channels = data.shape[0]

    # 2D or 3D
    is2d = len(input_shape) == 2

    # set step size to half of the window if necessary
    if step_size == None:
        if is2d:
            step_size = (1, input_shape[0] // 2, input_shape[1] // 2)
        else:
            step_size = (input_shape[0] // 2, input_shape[1] // 2, input_shape[2] // 2)

    # gaussian window for smooth block merging
    if is2d:
        g_window = gaussian_window((1, input_shape[0], input_shape[1]), sigma=input_shape[-1] / 4)
    else:
        g_window = gaussian_window(input_shape, sigma=input_shape[-1] / 4)

    # allocate space
    seg_cum = np.zeros(data.shape[1:])
    counts_cum = np.zeros(data.shape[1:])

    # define sliding window
    if is2d:
        sw = sliding_window_multichannel(data, step_size=step_size, window_size=(1, input_shape[0], input_shape[1]),
                                         track_progress=track_progress)
    else:
        sw = sliding_window_multichannel(data, step_size=step_size, window_size=input_shape,
                                         track_progress=track_progress)

    # start prediction
    batch_counter = 0
    if is2d:
        batch = np.zeros((batch_size, channels, input_shape[0], input_shape[1]))
    else:
        batch = np.zeros((batch_size, channels, input_shape[0], input_shape[1], input_shape[2]))
    positions = np.zeros((batch_size, 3), dtype=int)
    for (z, y, x, inputs) in sw:

        # fill batch
        batch[batch_counter, ...] = inputs
        positions[batch_counter, :] = [z, y, x]

        # increment batch counter
        batch_counter += 1

        # perform segmentation when a full batch is filled
        if batch_counter == batch_size:

            # convert to tensors
            if torch.cuda.is_available():
                inputs = torch.FloatTensor(batch).cuda()
            else:
                inputs = torch.FloatTensor(batch).cpu()

            # forward prop
            outputs = net(inputs)
            if type(outputs) is tuple:
                outputs = outputs[-1]
            outputs = F.softmax(outputs, dim=1)

            # cumulate segmentation volume
            for b in range(batch_size):
                (z_b, y_b, x_b) = positions[b, :]
                # take into account the gaussian filtering
                if is2d:
                    seg_cum[z_b:z_b + 1, y_b:y_b + input_shape[0], x_b:x_b + input_shape[1]] += \
                        np.multiply(g_window, outputs.data.cpu().numpy()[b, 1:2, :, :])
                    counts_cum[z_b:z_b + 1, y_b:y_b + input_shape[0], x_b:x_b + input_shape[1]] += g_window
                else:
                    seg_cum[z_b:z_b + input_shape[0], y_b:y_b + input_shape[1], x_b:x_b + input_shape[2]] += \
                        np.multiply(g_window, outputs.data.cpu().numpy()[b, 1, ...])
                    counts_cum[z_b:z_b + input_shape[0], y_b:y_b + input_shape[1], x_b:x_b + input_shape[2]] += g_window

            # reset batch counter
            batch_counter = 0

    # don't forget last batch
    # convert to tensors
    if torch.cuda.is_available():
        inputs = torch.FloatTensor(batch).cuda()
    else:
        inputs = torch.FloatTensor(batch).cpu()

    # forward prop
    outputs = net(inputs)
    if type(outputs) is tuple:
        outputs = outputs[-1]
    outputs = F.softmax(outputs, dim=1)

    # cumulate segmentation volume
    for b in range(batch_counter):
        (z_b, y_b, x_b) = positions[b, :]
        # take into account the gaussian filtering
        if is2d:
            seg_cum[z_b:z_b + 1, y_b:y_b + input_shape[0], x_b:x_b + input_shape[1]] += \
                np.multiply(g_window, outputs.data.cpu().numpy()[b, 1:2, :, :])
            counts_cum[z_b:z_b + 1, y_b:y_b + input_shape[0], x_b:x_b + input_shape[1]] += g_window
        else:
            seg_cum[z_b:z_b + input_shape[0], y_b:y_b + input_shape[1], x_b:x_b + input_shape[2]] += \
                np.multiply(g_window, outputs.data.cpu().numpy()[b, 1, ...])
            counts_cum[z_b:z_b + input_shape[0], y_b:y_b + input_shape[1], x_b:x_b + input_shape[2]] += g_window

    # crop out the symmetric extension and compute segmentation
    segmentation = np.divide(seg_cum, counts_cum)

    return segmentation


def segment(data, net, input_shape, batch_size=1, step_size=None, train=False, track_progress=False):
    """
    Segment a 3D image using a specific network
    :param data: 3D array (Z, Y, X) representing the 3D image
    :param net: image-to-image segmentation network
    :param input_shape: size of the inputs (either 2 or 3-tuple)
    :param batch_size: batch size for processing
    :param step_size: step size of the sliding window
    :param train: evaluate the network in training mode
    :param track_progress: optionally, for tracking progress with progress bar
    :return: the segmented image
    """

    return segment_multichannel(data[np.newaxis, ...], net, input_shape,
                                batch_size=batch_size, step_size=step_size, train=train, track_progress=track_progress)


def validate(net, data, labels, input_size, batch_size=1, write_dir=None, val_file=None,
             writer=None, epoch=0, track_progress=False):
    """
    Validate a network on a dataset and its labels
    :param net: image-to-image segmentation network
    :param data: 3D array (Z, Y, X) representing the 3D image
    :param labels: 3D array (Z, Y, X) representing the 3D labels
    :param input_size: size of the inputs (either 2 or 3-tuple) for processing
    :param batch_size: batch size for processing
    :param write_dir: optionally, specify a directory to write the output
    :param val_file: optionally, specify a file to write the validation results
    :param writer: optionally, summary writer for logging to tensorboard
    :param epoch: optionally, current epoch for logging to tensorboard
    :param track_progress: optionally, for tracking progress with progress bar
    :return: validation results, i.e. accuracy, precision, recall, f-score, jaccard and dice score
    """

    print('[%s] Validating the trained network' % (datetime.datetime.now()))

    if write_dir is not None and not os.path.exists(write_dir):
        os.mkdir(write_dir)

    segmentation = segment(data, net, input_size, batch_size=batch_size, track_progress=track_progress)
    j = jaccard(segmentation, labels)
    d = dice(segmentation, labels)
    a, p, r, f = accuracy_metrics(segmentation, labels)
    h = hausdorff_distance(segmentation, labels)[0]
    if write_dir is not None:
        print('[%s] Writing the output' % (datetime.datetime.now()))
        write_volume(255 * segmentation, write_dir, type='pngseq')
    if writer is not None:
        z = data.shape[0] // 2
        N = 1024
        if data.shape[1] > N:
            writer.add_image('val/input', data[z:z + 1, :N, :N], epoch)
            writer.add_image('val/segmentation', segmentation[z:z + 1, :N, :N], epoch)
        else:
            writer.add_image('val/input', data[z:z + 1, ...], epoch)
            writer.add_image('val/segmentation', segmentation[z:z + 1, ...], epoch)

    print('[%s] Network performance: Jaccard=%f - Dice=%f' % (datetime.datetime.now(), j, d))
    if val_file is not None:
        np.save(val_file, np.asarray([a, p, r, f, j, d, h]))
    return a, p, r, f, j, d, h



import cv2
import numpy as np
import numpy.random as rnd
import torch
import torch.nn.functional as F
from scipy.ndimage import spline_filter1d, zoom


class ToTensor(object):

    def __init__(self, cuda=True):
        """
        Transforms a numpy array into a tensor
        :param cuda: specifies whether the tensor should be transfered to the GPU or not
        """
        self.cuda = cuda

    def __call__(self, x):
        """
        Forward call
        :param x: input array (N_1, N_2, N_3, ...)
        :return: output tensor (N_1, N_2, N_3, ...)
        """
        if self.cuda:
            return torch.Tensor(x).cuda()
        else:
            return torch.Tensor(x)


class ToFloatTensor(object):

    def __init__(self, cuda=True):
        """
        Transforms a Tensor to a FloatTensor
        :param cuda: specifies whether the tensor should be transfered to the GPU or not
        """
        self.cuda = cuda

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (N_1, N_2, N_3, ...)
        :return: output tensor (N_1, N_2, N_3, ...)
        """
        if self.cuda:
            return x.float().cuda()
        else:
            return x.float()


class ToLongTensor(object):

    def __init__(self, cuda=True):
        """
        Transforms a Tensor to a LongTensor
        :param cuda: specifies whether the tensor should be transfered to the GPU or not
        """
        self.cuda = cuda

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (N_1, N_2, N_3, ...)
        :return: output tensor (N_1, N_2, N_3, ...)
        """
        if self.cuda:
            return x.long().cuda()
        else:
            return x.long()


class AddChannelAxis(object):

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (N_1, N_2, N_3, ...)
        :return: output tensor (1, N_1, N_2, N_3, ...)
        """
        return x.unsqueeze(0)


class AddNoise(object):

    def __init__(self, prob=0.5, sigma_min=0.0, sigma_max=1.0, include_segmentation=False,
                 include_weak_segmentation=False):
        """
        Adds noise to the input
        :param prob: probability of adding noise
        :param sigma_min: minimum noise standard deviation
        :param sigma_max: maximum noise standard deviation
        :param include_segmentation: 2nd half of the batch will not be augmented as this is assumed to be a (partial) segmentation
        """
        self.prob = prob
        self.sigma_min = sigma_min
        self.sigma_max = sigma_max
        self.include_segmentation = include_segmentation
        self.include_weak_segmentation = include_weak_segmentation

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, N_1, N_2, ...)
        :return: output tensor (B, N_1, N_2, ...)
        """

        if rnd.rand() < self.prob:
            sigma = rnd.uniform(self.sigma_min, self.sigma_max)
            if self.include_segmentation:
                sz = np.asarray(x.size())
                sz[0] = sz[0] // 2
                sz = tuple(sz)
                noise = torch.cat((torch.normal(0, sigma, sz), torch.zeros(sz)), dim=0)
            else:
                noise = torch.normal(0, sigma, x.size())
            if x.is_cuda:
                noise = noise.cuda()
            return x + noise
        else:
            return x


class Normalize(object):

    def __init__(self, mu=0, std=1):
        """
        Normalizes the input
        :param mu: mean of the normalization
        :param std: standard deviation of the normalization
        """
        self.mu = mu
        self.std = std

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (N_1, N_2, N_3, ...)
        :return: output tensor (N_1, N_2, N_3, ...)
        """
        return (x - self.mu) / self.std


class Scale(object):

    def __init__(self, scale_factor=(0.5, 1.5), mode='bilinear'):
        """
        Scales the input by a specific factor (randomly selected from a minimum-maximum range)
        :param scale_factor: minimum and maximum scaling factor
        """
        self.scale_factor = scale_factor
        self.mode = mode

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, [Z  , Y  ,] X)
        :return: output tensor (B, C, [Z' , Y' ,] X)
        """
        if type(self.scale_factor) == tuple:
            scale_factor = (self.scale_factor[1] - self.scale_factor[0]) * np.random.random_sample() + \
                           self.scale_factor[0]
        else:
            scale_factor = self.scale_factor
        return F.interpolate(x, scale_factor=scale_factor, mode=self.mode, align_corners=False)


class FlipX(object):

    def __init__(self, prob=1):
        """
        Perform a flip along the X axis
        :param prob: probability of flipping
        """
        self.prob = prob

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, N_1, N_2, ...)
        :return: output tensor (B, C, N_1, N_2, ...)
        """

        if rnd.rand() < self.prob:
            n = x.ndimension()
            return torch.flip(x, dims=[n - 1])
        else:
            return x


class FlipY(object):

    def __init__(self, prob=1):
        """
        Perform a flip along the Y axis
        :param prob: probability of flipping
        """
        self.prob = prob

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, N_1, N_2, ...)
        :return: output tensor (B, C, N_1, N_2, ...)
        """

        if rnd.rand() < self.prob:
            n = x.ndimension()
            return torch.flip(x, dims=[n - 2])
        else:
            return x


class FlipZ(object):

    def __init__(self, prob=1):
        """
        Perform a flip along the Z axis
        :param prob: probability of flipping
        """
        self.prob = prob

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, N_1, N_2, ...)
        :return: output tensor (B, C, N_1, N_2, ...)
        """

        if rnd.rand() < self.prob:
            n = x.ndimension()
            return torch.flip(x, dims=[n - 3])
        else:
            return x


class Rotate90(object):

    def __init__(self, prob=1):
        """
        Rotate the inputs by 90 degree angles
        :param prob: probability of rotating
        """
        self.prob = prob

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, N_1, N_2, ...)
        :return: output tensor (B, C, N_1, N_2, ...)
        """

        if rnd.rand() < self.prob:
            n = x.ndimension()
            return torch.rot90(x, k=rnd.randint(0, 4), dims=[n - 2, n - 1])
        else:
            return x


class RotateRandom_2D(object):

    def __init__(self, shape, prob=1.0, range_=200, cuda=True):
        """
        Rotate the inputs by a random amount of degrees within interval.
        :param shape: 2D shape of the input image
        :param range_: random degree interval size (symmetric around 0)
        :param cuda: specify whether the inputs are on the GPU.
        """
        self.shape = tuple(shape)
        self.cuda = cuda
        self.range = int(range_ / 2)
        self.prob = prob
        self.image_center = int(self.shape[0] / 2), int(self.shape[1] / 2)

        i = np.linspace(-1, 1, shape[0])
        j = np.linspace(-1, 1, shape[1])
        self.xv, self.yv = np.meshgrid(i, j)

    def _rotation_grid(self):
        angle = np.random.randint(-self.range, self.range)
        rot_matrix = cv2.getRotationMatrix2D(self.image_center, angle, 1.0)
        xv = cv2.warpAffine(self.xv, rot_matrix, self.xv.shape[1::-1], flags=cv2.INTER_CUBIC, borderValue=2)
        yv = cv2.warpAffine(self.yv, rot_matrix, self.yv.shape[1::-1], flags=cv2.INTER_CUBIC, borderValue=2)

        grid = torch.cat((torch.Tensor(xv).unsqueeze(-1), torch.Tensor(yv).unsqueeze(-1)), dim=-1)
        grid = grid.unsqueeze(0)
        if self.cuda:
            grid = grid.cuda()
        return grid

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, Y , X)
        :return: output tensor (B, C, Y , X)
        """

        if rnd.rand() < self.prob:
            grid = self._rotation_grid()
            grid = grid.repeat_interleave(x.size(0), dim=0)
            return F.grid_sample(x, grid)
        else:
            return x


class RotateRandom_3D(object):

    def __init__(self, shape, prob=1.0, range_=200, cuda=True):
        """
        Rotate the inputs by a random amount of degrees within interval.
        :param shape: 3D shape of the input image
        :param range_: random degree interval size (symmetric around 0)
        :param cuda: specify whether the inputs are on the GPU.
        """
        self.shape = tuple(shape)
        self.cuda = cuda
        self.range = int(range_ / 2)
        self.prob = prob
        self.image_center = int(self.shape[0] / 2), int(self.shape[1] / 2), int(self.shape[2] / 2)

        i = np.linspace(-1, 1, shape[0])
        j = np.linspace(-1, 1, shape[1])
        k = np.linspace(-1, 1, shape[2])
        self.xv, self.yv, self.zv = np.meshgrid(i, j, k)

    def _rotation_grid(self):
        angle = np.random.randint(-self.range, self.range)
        rot_matrix = cv2.getRotationMatrix2D(self.image_center, angle, 1.0)
        xv = cv2.warpAffine(self.xv, rot_matrix, self.xv.shape[1::-1], flags=cv2.INTER_CUBIC, borderValue=2)
        yv = cv2.warpAffine(self.yv, rot_matrix, self.yv.shape[1::-1], flags=cv2.INTER_CUBIC, borderValue=2)
        zv = cv2.warpAffine(self.zv, rot_matrix, self.zv.shape[1::-1], flags=cv2.INTER_CUBIC, borderValue=2)

        grid = torch.cat(
            (torch.Tensor(xv).unsqueeze(-1), torch.Tensor(yv).unsqueeze(-1), torch.Tensor(zv).unsqueeze(-1)), dim=-1)
        grid = grid.unsqueeze(0)
        if self.cuda:
            grid = grid.cuda()
        return grid

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, Z, Y , X)
        :return: output tensor (B, C, Z, Y , X)
        """

        if rnd.rand() < self.prob:
            grid = self._rotation_grid()
            grid = grid.repeat_interleave(x.size(0), dim=0)
            return F.grid_sample(x, grid)
        else:
            return x


class RandomCrop_2D(object):

    def __init__(self, crop_shape):
        """
        Selects a random crop from the input
        :param crop_shape: 2D shape of the crop
        """
        self.crop_shape = crop_shape

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, Y , X)
        :return: output tensor (B, C, Y', X')
        """
        r = np.random.randint(0, x.size(2) - self.crop_shape[0] + 1)
        c = np.random.randint(0, x.size(3) - self.crop_shape[1] + 1)
        return x[:, :, r:r + self.crop_shape[0], c:c + self.crop_shape[1]]


class RandomCrop_3D(object):

    def __init__(self, crop_shape):
        """
        Selects a random crop from the input
        :param crop_shape: 3D shape of the crop
        """
        self.crop_shape = crop_shape

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, Z , Y , X)
        :return: output tensor (B, C, Z', Y', X')
        """
        r = np.random.randint(0, x.size(2) - self.crop_shape[0] + 1)
        c = np.random.randint(0, x.size(3) - self.crop_shape[1] + 1)
        z = np.random.randint(0, x.size(4) - self.crop_shape[2] + 1)
        return x[:, :, r:r + self.crop_shape[0], c:c + self.crop_shape[1], z:z + self.crop_shape[2]]


class RandomDeformation_2D(object):

    def __init__(self, shape, prob=1, cuda=True, points=None, grid_size=8, sigma=0.01, n_grids=1000,
                 include_segmentation=False):
        """
        Apply random deformation to the inputs
        :param shape: shape of the inputs
        :param prob: probability of deforming the data
        :param cuda: specifies whether the inputs are on the GPU
        :param points: seed points for deformation
        :param grid_size: number of pixels between each grid point
        :param n_grids: number of grids to load in advance (chose more for higher variance in the data)
        :param include_segmentation: 2nd half of the batch needs casting to integers because of warping
        """
        self.shape = shape
        self.prob = prob
        self.cuda = cuda
        self.grid_size = grid_size
        if points == None:
            points = [shape[0] // self.grid_size, shape[1] // self.grid_size]
        self.points = points
        self.sigma = sigma
        self.n_grids = n_grids
        self.grids = []
        self.p = 10
        self.include_segmentation = include_segmentation

        i = np.linspace(-1, 1, shape[0])
        j = np.linspace(-1, 1, shape[1])
        xv, yv = np.meshgrid(i, j)

        grid = torch.cat((torch.Tensor(xv).unsqueeze(-1), torch.Tensor(yv).unsqueeze(-1)), dim=-1)
        grid = grid.unsqueeze(0)
        if cuda:
            grid = grid.cuda()
        self.grid = grid

        # generate several random grids in advance (less CPU work)
        for i in range(self.n_grids):
            self.grids.append(self._deformation_grid())

    def _deformation_grid(self):
        """
        Get a new random deformation grid
        :return: deformation tensor
        """
        sigma = np.random.rand() * self.sigma
        displacement = np.random.randn(*self.points, 2) * sigma

        # filter the displacement
        displacement_f = np.zeros_like(displacement)
        for d in range(0, displacement.ndim - 1):
            spline_filter1d(displacement, axis=d, order=3, output=displacement_f, mode='nearest')
            displacement = displacement_f

        # resample to proper size
        displacement_f = np.zeros((self.shape[0], self.shape[1], 2))
        for d in range(0, displacement.ndim - 1):
            displacement_f[:, :, d] = zoom(displacement[:, :, d], self.grid_size)

        displacement = torch.Tensor(displacement_f).unsqueeze(0)
        if self.cuda:
            displacement = displacement.cuda()
        grid = self.grid + displacement

        return grid

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, Y , X)
        :return: output tensor (B, C, Y , X)
        """

        if rnd.rand() < self.prob:
            grid = self.grids[rnd.randint(self.n_grids)]
            grid = grid.repeat_interleave(x.size(0), dim=0)
            x_aug = F.grid_sample(x, grid, padding_mode="border")
            if self.include_segmentation:
                x_aug[x.size(0) // 2:, ...] = torch.round(x_aug[x.size(0) // 2:, ...])
            return x_aug
        else:
            return x


class RandomDeformation_3D(object):

    def __init__(self, shape, prob=1, cuda=True, points=None, grid_size=8, sigma=0.01, n_grids=1000,
                 include_segmentation=False):
        """
        Apply random deformation to the inputs
        :param shape: shape of the inputs
        :param prob: probability of deforming the data
        :param cuda: specifies whether the inputs are on the GPU
        :param points: seed points for deformation
        :param grid_size: number of pixels between each grid point
        :param n_grids: number of grids to load in advance (chose more for higher variance in the data)
        :param include_segmentation: 2nd half of the batch needs casting to integers because of warping
        """
        self.shape = shape
        self.prob = prob
        self.cuda = cuda
        self.grid_size = grid_size
        if points == None:
            points = [shape[0] // self.grid_size, shape[1] // self.grid_size, shape[2] // self.grid_size]
        self.points = points
        self.sigma = sigma
        self.n_grids = n_grids
        self.grids = []
        self.p = 10
        self.include_segmentation = include_segmentation

        i = np.linspace(-1, 1, shape[0])
        j = np.linspace(-1, 1, shape[1])
        k = np.linspace(-1, 1, shape[2])
        xv, yv, zv = np.meshgrid(k, i, j)

        grid = torch.cat(
            (torch.Tensor(xv).unsqueeze(-1), torch.Tensor(yv).unsqueeze(-1), torch.Tensor(zv).unsqueeze(-1)), dim=-1)
        grid = grid.unsqueeze(0)
        if cuda:
            grid = grid.cuda()
        self.grid = grid

        # generate several random grids in advance (less CPU work)
        for i in range(self.n_grids):
            self.grids.append(self._deformation_grid())

    def _deformation_grid(self):
        """
        Get a new random deformation grid
        :return: deformation tensor
        """
        sigma = np.random.rand() * self.sigma
        displacement = np.random.randn(*self.points, 3) * sigma

        # filter the displacement
        displacement_f = np.zeros_like(displacement)
        for d in range(0, displacement.ndim - 1):
            spline_filter1d(displacement, axis=d, order=3, output=displacement_f, mode='nearest')
            displacement = displacement_f

        # resample to proper size
        displacement_f = np.zeros((self.shape[0], self.shape[1], self.shape[2], 3))
        for d in range(0, displacement.ndim - 1):
            displacement_f[:, :, :, d] = zoom(displacement[:, :, :, d], self.grid_size)

        displacement = torch.Tensor(displacement_f).unsqueeze(0)
        if self.cuda:
            displacement = displacement.cuda()
        grid = self.grid + displacement

        return grid

    def __call__(self, x):
        """
        Forward call
        :param x: input tensor (B, C, Z, Y , X)
        :return: output tensor (B, C, Z, Y , X)
        """

        if rnd.rand() < self.prob:
            grid = self.grids[rnd.randint(self.n_grids)]
            grid = grid.repeat_interleave(x.size(0), dim=0)
            x_aug = F.grid_sample(x, grid, padding_mode="border")
            if self.include_segmentation:
                x_aug[x.size(0) // 2:, ...] = torch.round(x_aug[x.size(0) // 2:, ...])
            return x_aug
        else:
            return x



import numpy as np
from scipy.spatial.distance import directed_hausdorff


def jaccard(x, y):
    """
    Jaccard index between two segmentations
    :param x: array
    :param y: array
    :return: the Jaccard index
    """

    # binarize
    x = x > 0.5
    y = y > 0.5

    # stabilizing constant
    eps = 1e-10

    # compute jaccard
    intersection = np.sum(np.multiply(x, y))
    union = np.sum(x) + np.sum(y) - intersection
    return (intersection + eps) / (union + eps)


def dice(x, y):
    """
    Dice coefficient between two segmentations
    :param x: array
    :param y: array
    :return: the Dice coefficient
    """

    # binarize
    x = x > 0.5
    y = y > 0.5

    # stabilizing constant
    eps = 1e-10

    # compute dice
    intersection = np.sum(np.multiply(x, y))
    return 2 * (intersection + eps) / (np.sum(x) + np.sum(y) + eps)


def accuracy_metrics(x, y):
    """
    Accuracy, precision, recall and f-score between two segmentations
    :param x: array
    :param y: array
    :return: the accuracy metrics
    """

    # binarize
    x = x > 0.5
    y = y > 0.5

    # stabilizing constant
    eps = 1e-10

    tp = np.sum(np.multiply(x, y))
    tn = np.sum(np.multiply(1 - x, 1 - y))
    fp = np.sum(np.multiply(x, 1 - y))
    fn = np.sum(np.multiply(1 - x, y))

    accuracy = (tp + tn + eps) / (tp + tn + fp + fn + eps)
    precision = (tp + eps) / (tp + fp + eps)
    recall = (tp + eps) / (tp + fn + eps)
    f_score = (2 * (precision * recall) + eps) / (precision + recall + eps)

    return accuracy, precision, recall, f_score


def hausdorff_distance(x, y):
    """
    Hausdorff distance between two segmentations
    :param x: array
    :param y: array
    :return: the hausdorff distance
    """

    # binarize
    x = x > 0.5
    y = y > 0.5

    hd_0 = 0
    hd_1 = 0
    hd = 0
    for i in range(x.shape[0]):
        hd_0 += directed_hausdorff(x[i, ...], y[i, ...])[0]
        hd_1 += directed_hausdorff(y[i, ...], x[i, ...])[0]
        hd += max(hd_0, hd_1)
    hd_0 /= x.shape[0]
    hd_1 /= x.shape[0]
    hd /= x.shape[0]

    return hd, hd_0, hd_1
